static const char *fsconfig_cmds[] = {
	[0] = "SET_FLAG",
	[1] = "SET_STRING",
	[2] = "SET_BINARY",
	[3] = "SET_PATH",
	[4] = "SET_PATH_EMPTY",
	[5] = "SET_FD",
	[6] = "CMD_CREATE",
	[7] = "CMD_RECONFIGURE",
	[8] = "CMD_CREATE_EXCL",
};
