/* Automatically generated by /bin/sh */
struct cmdname_help
{
    char name[16];
    char help[80];
};

static struct cmdname_help common_cmds[] = {
  {"annotate", "Read perf.data (created by perf record) and display annotated code"},
  {"archive", "Create archive with object files with build-ids found in perf.data file"},
  {"bench", "General framework for benchmark suites"},
  {"buildid-cache", "Manage build-id cache."},
  {"buildid-list", "List the buildids in a perf.data file"},
  {"c2c", "Shared Data C2C/HITM Analyzer."},
  {"config", "Get and set variables in a configuration file."},
  {"daemon", "Run record sessions on background"},
  {"data", "Data file related processing"},
  {"diff", "Read perf.data files and display the differential profile"},
  {"evlist", "List the event names in a perf.data file"},
  {"ftrace", "simple wrapper for kernel's ftrace functionality"},
  {"inject", "Filter to augment the events stream with additional information"},
  {"iostat", "Show I/O performance metrics"},
  {"kallsyms", "Searches running kernel for symbols"},
  {"kvm", "Tool to trace/measure kvm guest os"},
  {"list", "List all symbolic event types"},
  {"mem", "Profile memory accesses"},
  {"record", "Run a command and record its profile into perf.data"},
  {"report", "Read perf.data (created by perf record) and display the profile"},
  {"script", "Read perf.data (created by perf record) and display trace output"},
  {"stat", "Run a command and gather performance counter statistics"},
  {"test", "Runs sanity tests."},
  {"top", "System profiling tool."},
  {"version", "display the version of perf binary"},
#ifdef HAVE_LIBELF_SUPPORT
  {"probe", "Define new dynamic tracepoints"},
#endif /* HAVE_LIBELF_SUPPORT */
#if defined(HAVE_LIBTRACEEVENT)
  {"trace", "strace inspired tool"},
#endif /* HAVE_LIBTRACEEVENT */
#ifdef HAVE_LIBTRACEEVENT
  {"kmem", "Tool to trace/measure kernel memory properties"},
  {"kwork", "Tool to trace/measure kernel work properties (latencies)"},
  {"lock", "Analyze lock events"},
  {"sched", "Tool to trace/measure scheduler properties (latencies)"},
  {"timechart", "Tool to visualize total system behavior during a workload"},
#endif /* HAVE_LIBTRACEEVENT */
};
