static const char *kcmp_types[] = {
	[KCMP_FILE]	= "FILE",
	[KCMP_VM]	= "VM",
	[KCMP_FILES]	= "FILES",
	[KCMP_FS]	= "FS",
	[KCMP_SIGHAND]	= "SIGHAND",
	[KCMP_IO]	= "IO",
	[KCMP_SYSVSEM]	= "SYSVSEM",
	[KCMP_EPOLL_TFD]	= "EPOLL_TFD",
};
