static const char *fs_at_flags[] = {
	[ilog2(0x100) + 1] = "SYMLINK_NOFOLLOW",
	[ilog2(0x400) + 1] = "SYMLINK_FOLLOW",
	[ilog2(0x800) + 1] = "NO_AUTOMOUNT",
	[ilog2(0x1000) + 1] = "EMPTY_PATH",
	[ilog2(0x0000) + 1] = "STATX_SYNC_AS_STAT",
	[ilog2(0x2000) + 1] = "STATX_FORCE_SYNC",
	[ilog2(0x4000) + 1] = "STATX_DONT_SYNC",
	[ilog2(0x8000) + 1] = "RECURSIVE",
	[ilog2(0x0002) + 1] = "RENAME_EXCHANGE",
	[ilog2(0x0004) + 1] = "RENAME_WHITEOUT",
	[ilog2(0x200) + 1] = "REMOVEDIR",
	[ilog2(0x10000) + 1] = "EXECVE_CHECK",
};
